@@ -258,153 +258,165 @@ static const char *scope2str(unsigned char scope)
 	return "";
 }
 
-static bool get_ipv6_state(gchar *ifname)
+#define PROC_IPV4_CONF_PREFIX "/proc/sys/net/ipv4/conf"
+#define PROC_IPV6_CONF_PREFIX "/proc/sys/net/ipv6/conf"
+
+static int read_conf_value(const char *prefix, const char *ifname,
+					const char *suffix, int *value)
 {
-	int disabled;
 	gchar *path;
 	FILE *f;
-	bool enabled = false;
-
-	if (!ifname)
-		path = g_strdup("/proc/sys/net/ipv6/conf/all/disable_ipv6");
-	else
-		path = g_strdup_printf(
-			"/proc/sys/net/ipv6/conf/%s/disable_ipv6", ifname);
+	int err;
 
+	path = g_build_filename(prefix, ifname ? ifname : "all", suffix, NULL);
 	if (!path)
-		return enabled;
+		return -ENOMEM;
 
+	errno = 0;
 	f = fopen(path, "r");
+	if (!f) {
+		err = -errno;
+	} else {
+		errno = 0; /* Avoid stale errno values with fscanf */
+
+		err = fscanf(f, "%d", value);
+		if (err <= 0 && errno)
+			err = -errno;
+
+		fclose(f);
+	}
+
+	if (err <= 0)
+		connman_error("failed to read %s", path);
 
 	g_free(path);
 
-	if (f) {
-		if (fscanf(f, "%d", &disabled) > 0)
-			enabled = !disabled;
+	return err;
+}
+
+static int read_ipv4_conf_value(const char *ifname, const char *suffix,
+								int *value)
+{
+	return read_conf_value(PROC_IPV4_CONF_PREFIX, ifname, suffix, value);
+}
+
+static int read_ipv6_conf_value(const char *ifname, const char *suffix,
+								int *value)
+{
+	return read_conf_value(PROC_IPV6_CONF_PREFIX, ifname, suffix, value);
+}
+
+static int write_conf_value(const char *prefix, const char *ifname,
+					const char *suffix, int value) {
+	gchar *path;
+	FILE *f;
+	int rval;
+
+	path = g_build_filename(prefix, ifname ? ifname : "all", suffix, NULL);
+	if (!path)
+		return -ENOMEM;
+
+	f = fopen(path, "r+");
+	if (!f) {
+		rval = -errno;
+	} else {
+		rval = fprintf(f, "%d", value);
 		fclose(f);
 	}
 
+	if (rval <= 0)
+		connman_error("failed to set %s value %d", path, value);
+
+	g_free(path);
+
+	return rval;
+}
+
+static int write_ipv4_conf_value(const char *ifname, const char *suffix,
+								int value)
+{
+	return write_conf_value(PROC_IPV4_CONF_PREFIX, ifname, suffix, value);
+}
+
+static int write_ipv6_conf_value(const char *ifname, const char *suffix,
+								int value)
+{
+	return write_conf_value(PROC_IPV6_CONF_PREFIX, ifname, suffix, value);
+}
+
+static bool get_ipv6_state(gchar *ifname)
+{
+	int disabled;
+	bool enabled = false;
+
+	if (read_ipv6_conf_value(ifname, "disable_ipv6", &disabled) > 0)
+		enabled = !disabled;
+
 	return enabled;
 }
 
-static void set_ipv6_state(gchar *ifname, bool enable)
+static int set_ipv6_state(gchar *ifname, bool enable)
 {
-	gchar *path;
-	FILE *f;
+	int disabled = enable ? 0 : 1;
 
-	if (!ifname)
-		path = g_strdup("/proc/sys/net/ipv6/conf/all/disable_ipv6");
-	else
-		path = g_strdup_printf(
-			"/proc/sys/net/ipv6/conf/%s/disable_ipv6", ifname);
+	DBG("%s %d", ifname, disabled);
 
-	if (!path)
-		return;
-
-	f = fopen(path, "r+");
-
-	g_free(path);
-
-	if (!f)
-		return;
-
-	if (!enable)
-		fprintf(f, "1");
-	else
-		fprintf(f, "0");
-
-	fclose(f);
+	return write_ipv6_conf_value(ifname, "disable_ipv6", disabled);
 }
 
 static int get_ipv6_privacy(gchar *ifname)
 {
-	gchar *path;
-	FILE *f;
 	int value;
 
 	if (!ifname)
 		return 0;
 
-	path = g_strdup_printf("/proc/sys/net/ipv6/conf/%s/use_tempaddr",
-								ifname);
-
-	if (!path)
-		return 0;
-
-	f = fopen(path, "r");
-
-	g_free(path);
-
-	if (!f)
-		return 0;
-
-	if (fscanf(f, "%d", &value) <= 0)
+	if (read_ipv6_conf_value(ifname, "use_tempaddr", &value) < 0)
 		value = 0;
 
-	fclose(f);
-
 	return value;
 }
 
 /* Enable the IPv6 privacy extension for stateless address autoconfiguration.
  * The privacy extension is described in RFC 3041 and RFC 4941
  */
-static void set_ipv6_privacy(gchar *ifname, int value)
+static int set_ipv6_privacy(gchar *ifname, int value)
 {
-	gchar *path;
-	FILE *f;
-
 	if (!ifname)
-		return;
-
-	path = g_strdup_printf("/proc/sys/net/ipv6/conf/%s/use_tempaddr",
-								ifname);
-
-	if (!path)
-		return;
+		return -EINVAL;
 
 	if (value < 0)
 		value = 0;
 
-	f = fopen(path, "r+");
-
-	g_free(path);
-
-	if (!f)
-		return;
-
-	fprintf(f, "%d", value);
-	fclose(f);
+	return write_ipv6_conf_value(ifname, "use_tempaddr", value);
 }
 
 static int get_rp_filter(void)
 {
-	FILE *f;
-	int value = -EINVAL, tmp;
+	int value;
 
-	f = fopen("/proc/sys/net/ipv4/conf/all/rp_filter", "r");
-
-	if (f) {
-		if (fscanf(f, "%d", &tmp) == 1)
-			value = tmp;
-		fclose(f);
-	}
+	if (read_ipv4_conf_value(NULL, "rp_filter", &value) < 0)
+		value = -EINVAL;
 
 	return value;
 }
 
-static void set_rp_filter(int value)
+static int set_rp_filter(int value)
 {
-	FILE *f;
+	/* 0 = no validation, 1 = strict mode, 2 = loose mode */
+	switch (value) {
+	case -1:
+		value = 0;
+		/* fall through */
+	case 0:
+	case 1:
+	case 2:
+		break;
+	default:
+		return -EINVAL;
+	}
 
-	f = fopen("/proc/sys/net/ipv4/conf/all/rp_filter", "r+");
-
-	if (!f)
-		return;
-
-	fprintf(f, "%d", value);
-
-	fclose(f);
+	return write_ipv4_conf_value(NULL, "rp_filter", value);
 }
 
 int __connman_ipconfig_set_rp_filter()
@@ -696,6 +708,25 @@ static inline gint check_duplicate_address(gconstpointer a, gconstpointer b)
 	return g_strcmp0(addr1->local, addr2->local);
 }
 
+static bool is_index_p2p_service(int index)
+{
+	struct connman_service *service;
+	enum connman_service_type type;
+
+	service = __connman_service_lookup_from_index(index);
+	if (!service)
+		return false;
+
+	type = connman_service_get_type(service);
+	switch (type) {
+	case CONNMAN_SERVICE_TYPE_P2P:
+	case CONNMAN_SERVICE_TYPE_VPN:
+		return true;
+	default:
+		return false;
+	}
+}
+
 int __connman_ipconfig_newaddr(int index, int family, const char *label,
 				unsigned char prefixlen, const char *address)
 {
@@ -718,6 +749,9 @@ int __connman_ipconfig_newaddr(int index, int family, const char *label,
 	ipaddress->prefixlen = prefixlen;
 	ipaddress->local = g_strdup(address);
 
+	if (is_index_p2p_service(index))
+		connman_ipaddress_set_p2p(ipaddress, true);
+
 	if (g_slist_find_custom(ipdevice->address_list, ipaddress,
 					check_duplicate_address)) {
 		connman_ipaddress_free(ipaddress);
@@ -1186,6 +1220,15 @@ void __connman_ipconfig_set_prefixlen(struct connman_ipconfig *ipconfig,
 	ipconfig->address->prefixlen = prefixlen;
 }
 
+static void ipconfig_set_p2p(int index, struct connman_ipconfig *ipconfig)
+{
+	if (!is_index_p2p_service(index))
+		return;
+
+	connman_ipaddress_set_p2p(ipconfig->address, true);
+	connman_ipaddress_set_p2p(ipconfig->system, true);
+}
+
 static struct connman_ipconfig *create_ipv6config(int index)
 {
 	struct connman_ipconfig *ipv6config;
@@ -1217,6 +1260,8 @@ static struct connman_ipconfig *create_ipv6config(int index)
 
 	ipv6config->system = connman_ipaddress_alloc(AF_INET6);
 
+	ipconfig_set_p2p(index, ipv6config);
+
 	DBG("ipconfig %p index %d method %s", ipv6config, index,
 		__connman_ipconfig_method2string(ipv6config->method));
 
@@ -1255,6 +1300,8 @@ struct connman_ipconfig *__connman_ipconfig_create(int index,
 
 	ipconfig->system = connman_ipaddress_alloc(AF_INET);
 
+	ipconfig_set_p2p(index, ipconfig);
+
 	DBG("ipconfig %p index %d", ipconfig, index);
 
 	return ipconfig;
@@ -1451,10 +1498,8 @@ int __connman_ipconfig_address_unset(struct connman_ipconfig *ipconfig)
 			err = connman_inet_clear_address(ipconfig->index,
 							ipconfig->address);
 		else if (ipconfig->type == CONNMAN_IPCONFIG_TYPE_IPV6)
-			err = connman_inet_clear_ipv6_address(
-						ipconfig->index,
-						ipconfig->address->local,
-						ipconfig->address->prefixlen);
+			err = connman_inet_clear_ipv6_address(ipconfig->index,
+							ipconfig->address);
 		else
 			err = -EINVAL;
 
@@ -1647,6 +1692,9 @@ int __connman_ipconfig_enable(struct connman_ipconfig *ipconfig)
 		connman_ipaddress_clear(ipdevice->config_ipv4->system);
 
 		__connman_ipconfig_unref(ipdevice->config_ipv4);
+
+		g_free(ipdevice->ipv4_gateway);
+		ipdevice->ipv4_gateway = NULL;
 	}
 
 	if (type == CONNMAN_IPCONFIG_TYPE_IPV6 &&
@@ -1657,6 +1705,9 @@ int __connman_ipconfig_enable(struct connman_ipconfig *ipconfig)
 		connman_ipaddress_clear(ipdevice->config_ipv6->system);
 
 		__connman_ipconfig_unref(ipdevice->config_ipv6);
+
+		g_free(ipdevice->ipv6_gateway);
+		ipdevice->ipv6_gateway = NULL;
 	}
 
 	if (type == CONNMAN_IPCONFIG_TYPE_IPV4)
@@ -1719,6 +1770,10 @@ int __connman_ipconfig_disable(struct connman_ipconfig *ipconfig)
 		connman_ipaddress_clear(ipdevice->config_ipv4->system);
 		__connman_ipconfig_unref(ipdevice->config_ipv4);
 		ipdevice->config_ipv4 = NULL;
+
+		g_free(ipdevice->ipv4_gateway);
+		ipdevice->ipv4_gateway = NULL;
+
 		return 0;
 	}
 
@@ -1728,6 +1783,10 @@ int __connman_ipconfig_disable(struct connman_ipconfig *ipconfig)
 		connman_ipaddress_clear(ipdevice->config_ipv6->system);
 		__connman_ipconfig_unref(ipdevice->config_ipv6);
 		ipdevice->config_ipv6 = NULL;
+
+		g_free(ipdevice->ipv6_gateway);
+		ipdevice->ipv6_gateway = NULL;
+
 		return 0;
 	}
 
